// Spanning.c
// Studying spanning arangements as proposed by Werner Keym and Valery Liskovets (see README)
// Alain Brobecker - 2018/08
#include <stdio.h>  // file functions, etc...
#include <stdlib.h> // exit(), malloc()
#include <string.h> // strlen()
#define TRUE 1
#define FALSE 0

#define FS -1
#define ES  0
#define WK  1
#define WQ  2
#define WR  3
#define WB  4
#define WS  5 //Springer
#define WP  6
#define BK  7
#define BQ  8
#define BR  9
#define BB 10
#define BS 11 //Springer
#define BP 12
#define WN 13 //NightRider
#define BN 14
#define WO 15 //Illegal pawn
#define BO 16

#define MaxNbPieces 16
#define Border 2        //enough for S and N
#define MaxSize 16
#define MinSize 3
int Width=8;
int Height=8;

char Board[(MaxSize+2*Border)*(MaxSize+2*Border)];
char ControlBoard[MaxNbPieces+1][(MaxSize+2*Border)*(MaxSize+2*Border)];
int piece[MaxNbPieces],pieceX[MaxNbPieces],pieceY[MaxNbPieces];
int NbPieces;
int NbSolutions;

//**** READNUMBER ****
int ReadNumber(s) char *s; {
  int n,t,sign;
  unsigned char c;
  n=0;
  sign=1;
  t=0;
  c=s[t];
  if(c=='-') { sign=-1; t++; c=s[t]; }
  while(c!=0) {
    if((c<'0') || (c>'9')) {
      printf("Error: %s should be a number",s);
      exit(1);
    }
    n=n*10+c-'0';
    t++;
    c=s[t];
  }
  return(n*sign);
}

//**** PrintOnePiece ****
void PrintOnePiece(int piece) {
  switch(piece) {
    case WK: printf("K"); break; case BK: printf("k"); break;
    case WQ: printf("Q"); break; case BQ: printf("q"); break;
    case WR: printf("R"); break; case BR: printf("r"); break;
    case WB: printf("B"); break; case BB: printf("b"); break;
    case WS: printf("S"); break; case BS: printf("s"); break;
    case WP: printf("P"); break; case BP: printf("p"); break;
    case WN: printf("N"); break; case BN: printf("n"); break; //NightRider
    case WO: printf("O"); break; case BO: printf("o"); break; //Illegal pawn
    case FS: printf("x"); break; //Hole
    default: printf("?"); break; // error
  }
}


//**** PrintFEN ****
int PF_x,PF_y,PF_pos,PF_nb;
void PrintFEN() {
  for(PF_y=0;PF_y<Height;PF_y++) {
    PF_nb=0;
    for(PF_x=0;PF_x<Width;PF_x++) {
      PF_pos=(PF_y+Border)*(Width+2*Border)+PF_x+Border;
      if(Board[PF_pos]==ES) {
        PF_nb++;
      } else {
        if(PF_nb!=0) { printf("%d",PF_nb); PF_nb=0; }
        PrintOnePiece(Board[PF_pos]);
      }
      PF_pos++;
    }
    if(PF_nb!=0) { printf("%d",PF_nb); }
    if(PF_y!=Height-1) { printf("/"); }
  }
}

//**** PrintPieces ****
void PrintPieces() {
  int x,y,c;
  for(c=0;c<NbPieces;c++) {
  x=pieceX[c];
  y=pieceY[c];
  PrintOnePiece(piece[c]);
  printf("%c%d",'a'+x,Height-y);
  if( ((x + y) & 1)==1 ) { printf(";1;"); } else {printf(";0;"); }//dark square?
  }
}

//**** LeaperControl ****
int LeaperControl(int depth,int x,int y,int dx,int dy,int attack) {
  int pos;
  pos=(y+dy+Border)*(Width+2*Border)+x+dx+Border;
  if((Board[pos]==ES) || (Board[pos]==FS)) {
    ControlBoard[depth+1][pos]+=1;
    return(attack);
  } else {
    return(TRUE);
  }
}

//**** RiderControl ****
int RiderControl(int depth,int x,int y,int dx,int dy,int attack) {
  int pos;
  pos=(y+dy+Border)*(Width+2*Border)+x+dx+Border;
  while(Board[pos]==ES) {
    ControlBoard[depth+1][pos]+=1;
    pos+=dy*(Width+2*Border)+dx;
  }
  //ended by reaching the border or a piece?
  if(Board[pos]==FS) {
    return(attack);
  } else {
    return(TRUE);
  }
}

//**** CountNbSquaresAttacked() ****
int MinNbAttacks,MaxNbAttacks;
int CountNbSquaresAttacked(int depth) {
  int x,y,pos,c,NbSquaresAttacked;
  NbSquaresAttacked=0;
  MinNbAttacks=NbPieces+1;
  MaxNbAttacks=0;
  for(y=0;y<Height;y++) {
    for(x=0;x<Width;x++) {
      pos=(y+Border)*(Width+2*Border)+x+Border;
      c=ControlBoard[depth+1][pos];
      if(c!=0) {
        NbSquaresAttacked++;
        if(MinNbAttacks>c) { MinNbAttacks=c; }
        if(MaxNbAttacks<c) { MaxNbAttacks=c; }
      }
    }//next x
  }//next y
  return(NbSquaresAttacked);
}

//**** CheckForSquareColor ****
int NbPiecesOnDarkSquares,NbBishopsOnDarkSquares;
void CheckForSquareColor() {
  int x,y,pos;
  NbPiecesOnDarkSquares=0;
  NbBishopsOnDarkSquares=0;
  for(y=0;y<Height;y++) {
    for(x=0;x<Width;x++) {
      if( ((x + y) & 1)==1 ) { //dark square?
        pos=(y+Border)*(Width+2*Border)+x+Border;
        switch(Board[pos]) {
          case WB: case BB: NbBishopsOnDarkSquares++; break;
      case ES: case FS: break;
          default: NbPiecesOnDarkSquares++; break;
        }
      }
    }//next x
  }//next y
  NbPiecesOnDarkSquares+=NbBishopsOnDarkSquares;
}

//**** RecursivelyPlaceOnePiece ****
void RecursivelyPlaceOnePiece(int depth) {
  int x,y,startx,starty,pos,c,AttackAnother,CurrentPiece;
  startx=0;
  starty=0;
  //if same piece as previous one, start after it
  if((depth!=0) && (piece[depth]==piece[depth-1])) {
    startx=pieceX[depth-1]+1;
    starty=pieceY[depth-1];
    if(startx==Width) {
      starty+=1;
      startx=0;
    }
  }
  //try to put piece on all ES square
  for(y=starty;y<Height;y++) {
    if(y!=starty) { startx=0; }
    for(x=startx;x<Width;x++) {
      pos=(y+Border)*(Width+2*Border)+x+Border;
    CurrentPiece=piece[depth];
      //square is empty and not controlled?
      if(    (Board[pos]==ES) //square is empty?
          && (ControlBoard[depth][pos]==0) //and not controlled?
//!!!BUG          && ((CurrentPiece!=WP) || (y!=0) || (y!=Height-1)) //and not a WP on 1st or last rank?
//!!!BUG          && ((CurrentPiece!=BP) || (y!=0) || (y!=Height-1)) ) { //and not a BP on 1st or last rank?
          && ((CurrentPiece!=WP) || (y!=0)) //and not a WP on 1st rank
          && ((CurrentPiece!=WP) || (y!=Height-1)) //and not a WP on or last rank?
          && ((CurrentPiece!=BP) || (y!=0)) //and not a BP on 1st rank
          && ((CurrentPiece!=BP) || (y!=Height-1)) ) { //and not a BP on or last rank?
        //put piece
        Board[pos]=CurrentPiece;
        pieceX[depth]=x;
        pieceY[depth]=y;
        //copy ControlBoard
        for(c=0;c<(Width+2*Border)*(Height+2*Border);c++) { ControlBoard[depth+1][c]=ControlBoard[depth][c]; }
        AttackAnother=FALSE;
        switch(CurrentPiece) {
          case WK: case BK:
            AttackAnother=LeaperControl(depth,x,y, 1,-1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1, 0,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 0, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-1, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-1, 0,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-1,-1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 0,-1,AttackAnother);
            break;
          case WQ: case BQ:
            AttackAnother=RiderControl(depth,x,y, 1, 0,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1, 0,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 0, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 0,-1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 1, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 1,-1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1,-1,AttackAnother);
            break;
          case WR: case BR:
            AttackAnother=RiderControl(depth,x,y, 1, 0,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1, 0,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 0, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 0,-1,AttackAnother);
            break;
          case WB: case BB:
            AttackAnother=RiderControl(depth,x,y, 1, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 1,-1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1,-1,AttackAnother);
            break;
          case WS: case BS:
            AttackAnother=LeaperControl(depth,x,y, 1, 2,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1,-2,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-1, 2,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-1,-2,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 2, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 2,-1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-2, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y,-2,-1,AttackAnother);
            break;
          case WP:
            AttackAnother=LeaperControl(depth,x,y,-1,-1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1,-1,AttackAnother);
            break;
          case BP:
            AttackAnother=LeaperControl(depth,x,y,-1, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1, 1,AttackAnother);
            break;
          case WN: case BN: //NightRider
            AttackAnother=RiderControl(depth,x,y, 1, 2,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 1,-2,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1, 2,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-1,-2,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 2, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y, 2,-1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-2, 1,AttackAnother);
            AttackAnother=RiderControl(depth,x,y,-2,-1,AttackAnother);
            break;
          case WO:
            AttackAnother=LeaperControl(depth,x,y,-1,-1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1,-1,AttackAnother);
            break;
          case BO:
            AttackAnother=LeaperControl(depth,x,y,-1, 1,AttackAnother);
            AttackAnother=LeaperControl(depth,x,y, 1, 1,AttackAnother);
            break;
          //Holes do not attack
        }
        //no piece was attacked, so go on
        if(AttackAnother==FALSE) {
          //process if last piece was put
          if(depth==NbPieces-1) {
            if(CountNbSquaresAttacked(depth)==Width*Height-NbPieces) { //All other squares attacked?
              CheckForSquareColor();
              printf("%d;%d;%d;",MinNbAttacks,MaxNbAttacks,NbPiecesOnDarkSquares);
              PrintPieces();
              PrintFEN();
              printf("\n");
              NbSolutions++;
            }
          } else {
            RecursivelyPlaceOnePiece(depth+1);
          }
        }
        //remove piece
        Board[pos]=ES;
      }
    }//next x
  }//next y
}



//**** MAIN ****
int main(argc,argv) int argc; char *argv[]; {
  setbuf(stdout, NULL); // no buffering for output!
  int x,y,pos,c;
  //Read from command line
  if(argc<=1) { printf("Usage: %s Pieces [Width] [Height]\n       %s KQRRBBSS\n       %s KQRBSP 6 6",argv[0],argv[0],argv[0]); exit(1); }
  NbPieces=strlen(argv[1]);
  for(c=0;c<NbPieces;c++) {
    switch(argv[1][c]) {
      case 'K': piece[c]=WK; break;
      case 'Q': piece[c]=WQ; break;
      case 'R': piece[c]=WR; break;
      case 'B': piece[c]=WB; break;
      case 'S': piece[c]=WS; break;
      case 'P': piece[c]=WP; break;
      case 'k': piece[c]=BK; break;
      case 'q': piece[c]=BQ; break;
      case 'r': piece[c]=BR; break;
      case 'b': piece[c]=BB; break;
      case 's': piece[c]=BS; break;
      case 'p': piece[c]=BP; break;
      case 'N': piece[c]=WN; break; //NightRider
      case 'n': piece[c]=BN; break;
      case 'O': piece[c]=WO; break; //Illegal pawn
      case 'o': piece[c]=BO; break;
      case 'H': piece[c]=FS; break; //Hole
      default: printf("%c is not a valid piece\n",argv[1][c]); exit(-1);
    }
  }
  //sort pieces in increasing order to have Holes=FS=-1 at the beginning and similar pieces nearby
  for(x=0;x<NbPieces-1;x++) {
    for(y=x+1;y<NbPieces;y++) {
      if(piece[y]<piece[x]) {
        c=piece[x];
        piece[x]=piece[y];
        piece[y]=c;
      }
    }//next y
  }//next x

  //optionnally modify Width and Height
  if(argc>2) { Width=ReadNumber(argv[2]); if((Width>MaxSize) || (Width<MinSize)) { printf("We must have %d <= Width <=%d\n",MinSize,MaxSize); exit(-1); } }
  if(argc>3) { Height=ReadNumber(argv[3]); if((Height>MaxSize) || (Height<MinSize)) { printf("We must have %d <= Height <=%d\n",MinSize,MaxSize); exit(-1); } }

  //initialise boards
  for(pos=0;pos<(Width+2*Border)*(Height+2*Border);pos++) { Board[pos]=FS; ControlBoard[0][pos]=0; }
  for(y=0;y<Height;y++) {
    for(x=0;x<Width;x++) {
      pos=(y+Border)*(Width+2*Border)+x+Border;
      Board[pos]=ES;
    }//next x
  }//next y

  printf("MinNbAttacks;MaxNbAttacks;NbPiecesOnDarkSquares;");
  for(c=0;c<NbPieces;c++) { PrintOnePiece(piece[c]); printf(";d");PrintOnePiece(piece[c]); printf(";"); }
  printf("FEN\n");

  NbSolutions=0;
  RecursivelyPlaceOnePiece(0);
  printf("%d solutions\n",NbSolutions);
  exit(0);
}
